# Copyright 2020 Axis Communications AB.
#
# For a full list of individual contributors, please see the commit history.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""IUT data structure module."""
import os
from packageurl import PackageURL


class Iut:  # pylint: disable=too-few-public-methods
    """Data object for IUTs."""

    def __init__(self, product):
        """Initialize.

        :param product: Dictionary to set attributes from.
                        Should be the response from pool plugin list.
        :type product: dict
        """
        try:
            self.load_environment(product.pop("environment"))
        except KeyError:
            pass
        try:
            self.prepare(product.pop("commands"))
        except KeyError:
            pass
        product["identity"] = PackageURL.from_string(product["identity"])
        for key, value in product.items():
            setattr(self, key, value)
        self._product_dict = product

    @staticmethod
    def load_environment(environment):
        """Load and set environment variables from IUT definition.

        :param environment: Environment variables to set.
        :type environment: dict
        """
        for key, value in environment.items():
            os.environ[key] = value

    def prepare(self, commands):
        """Prepare IUT and IUT environment for tests.

        :param commands: Commands to execute.
        :type commands: list
        """

    @property
    def as_dict(self):
        """Return IUT as a dictionary."""
        return self._product_dict

    def __repr__(self):
        """Represent IUT as string."""
        try:
            return self._product_dict.get("identity").to_string()
        except:  # noqa pylint:disable=bare-except
            return "Unknown"



# Copyright 2020 Axis Communications AB.
#
# For a full list of individual contributors, please see the commit history.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""ETR executor module."""
import os
import shlex
import logging
import signal
from pprint import pprint
from tempfile import mkdtemp
from etos_test_runner.test_regex import TEST_REGEX

BASE = os.path.dirname(os.path.abspath(__file__))


class TestCheckoutTimeout(TimeoutError):
    """Test checkout timeout exception."""


def _test_checkout_signal_handler(signum, frame):  # pylint:disable=unused-argument
    """Raise timeout error on test checkout."""
    raise TestCheckoutTimeout("Took too long to checkout test cases.")


class Executor:  # pylint: disable=too-many-instance-attributes
    """Execute a single test-case, -class, -module, -folder etc."""

    report_path = "test_output.log"
    test_name = ""
    current_test = None
    test_checkouts = {}
    logger = logging.getLogger("Executor")

    # pylint: disable=too-many-arguments
    def __init__(self, test, iut, etos):
        """Initialize.

        :param test: Test to execute.
        :type test: str
        :param iut: IUT to execute test on.
        :type iut: :obj:`etr.lib.iut.Iut`
        """
        self.test = test
        self.tests = {}

        self.test_environment_variables = {}
        self.test_command = None
        self.pre_test_execution = []
        self.test_command_input_arguments = {}
        self.checkout_command = []

        self.constraints = test.get("constraints", [])
        for constraint in self.constraints:
            if constraint.get("key") == "ENVIRONMENT":
                self.test_environment_variables = constraint.get("value")
            elif constraint.get("key") == "COMMAND":
                self.test_command = constraint.get("value")
            elif constraint.get("key") == "EXECUTE":
                self.pre_test_execution = constraint.get("value")
            elif constraint.get("key") == "PARAMETERS":
                self.test_command_input_arguments = constraint.get("value")
            elif constraint.get("key") == "CHECKOUT":
                self.checkout_command = constraint.get("value")

        self.test_name = test.get("testCase").get("id")
        self.test_id = test.get("id")
        self.iut = iut
        self.etos = etos
        self.context = self.etos.config.get("context")
        self.result = True

    def _checkout_tests(self, test_checkout):
        """Check out tests for this execution.

        :param test_checkout: Test checkout parameters from test suite.
        :type test_checkout: list
        """
        with open(os.path.join(BASE, "checkout.sh"), "w") as checkout_file:
            checkout_file.write('eval "$(pyenv init -)"\n')
            checkout_file.write("pyenv shell --unset\n")
            for command in test_checkout:
                checkout_file.write("{} || exit 1\n".format(command))

        signal.signal(signal.SIGALRM, _test_checkout_signal_handler)
        signal.alarm(60)
        try:
            checkout = os.path.join(BASE, "checkout.sh")
            success, output = self.etos.utils.call(
                ["/bin/bash", checkout], shell=True, wait_output=False
            )
        finally:
            signal.alarm(0)
        if not success:
            pprint(output)
            raise Exception("Could not checkout tests using '{}'".format(test_checkout))

    def test_directory(self, test_checkout):
        """Test directory for the test checkout.

        If a test directory does not already exist, generate it by calling the
        supplied command from test suite.
        If it does exist, just return that directory.

        :param test_checkout: Test checkout parameters from test suite.
        :type test_checkout: list
        :return: Folder where to execute a testcase
        :rtype: str
        """
        string_checkout = " ".join(test_checkout)
        if self.test_checkouts.get(string_checkout) is None:
            test_folder = mkdtemp(dir=os.getcwd())
            with self.etos.utils.chdir(test_folder):
                self._checkout_tests(test_checkout)
            self.test_checkouts[string_checkout] = test_folder
        return self.test_checkouts.get(string_checkout)

    def _build_test_command(self):
        """Build up the actual test command based on data from event."""
        executor = os.path.join(BASE, "executor.sh")
        test_command = ""
        parameters = []

        for parameter, value in self.test_command_input_arguments.items():
            if value == "":
                parameters.append(parameter)
            else:
                parameters.append("{}={}".format(parameter, value))

        test_command = "{} {} {} 2>&1".format(
            executor, self.test_command, " ".join(parameters)
        )
        return test_command

    def __enter__(self):
        """Enter context."""
        return self

    def __exit__(self, _type, value, traceback):
        """Exit context."""

    @staticmethod
    def _pre_execution(command):
        """Write pre execution command to a shell script.

        :param command: Environment and pre execution shell command to write to shell script.
        :type command: str
        """
        with open(os.path.join(BASE, "environ.sh"), "w") as environ_file:
            for arg in command:
                environ_file.write("{} || exit 1\n".format(arg))

    def _build_environment_command(self):
        """Build command for setting environment variables prior to execution.

        :return: Command to run pre execution.
        :rtype: str
        """
        environments = [
            "export {}={}".format(key, shlex.quote(value))
            for key, value in self.test_environment_variables.items()
        ]
        return environments + self.pre_test_execution

    def _triggered(self, test_name):
        """Send a test case triggered event.

        :param test_name: Name of test that is triggered.
        :type test_name: str
        :return: Test case triggered event created and sent.
        :rtype: :obj:`eiffellib.events.eiffel_test_case_triggered_event.EiffelTestCaseTriggeredEvent`  # pylint:disable=line-too-long
        """
        return self.etos.events.send_test_case_triggered(
            {"id": test_name}, self.iut.artifact, links={"CONTEXT": self.context}
        )

    def _started(self, test_name):
        """Send a testcase started event.

        :param test_name: Name of test that has started.
        :type test_name: str
        :return: Test case started event created and sent.
        :rtype: :obj:`eiffellib.events.eiffel_test_case_started_event.EiffelTestCaseStartedEvent`
        """
        triggered = self.tests[test_name].get("triggered")
        if triggered is None:
            return None
        return self.etos.events.send_test_case_started(
            triggered, links={"CONTEXT": self.context}
        )

    def _finished(self, test_name, result):
        """Send a testcase finished event.

        :param test_name: Name of test that is finished.
        :type test_name: str
        :param result: Result of test case.
        :type result: str
        :return: Test case finished event created and sent.
        :rtype: :obj:`eiffellib.events.eiffel_test_case_finished_event.EiffelTestCaseFinishedEvent`
        """
        triggered = self.tests[test_name].get("triggered")
        if triggered is None:
            return None

        if result == "ERROR":
            outcome = {"verdict": "FAILED", "conclusion": "INCONCLUSIVE"}
        elif result == "FAILED":
            outcome = {"verdict": "FAILED", "conclusion": "FAILED"}
        elif result == "SKIPPED":
            outcome = {
                "verdict": "PASSED",
                "conclusion": "SUCCESSFUL",
                "description": "SKIPPED",
            }
        else:
            outcome = {"verdict": "PASSED", "conclusion": "SUCCESSFUL"}
        return self.etos.events.send_test_case_finished(
            triggered, outcome, links={"CONTEXT": self.context}
        )

    def parse(self, line):
        """Parse test output in order to send test case events.

        :param line: Line to parse.
        :type line: str
        """
        if not isinstance(line, str):
            return
        test_name = TEST_REGEX["test_name"].findall(line)
        if test_name:
            self.current_test = test_name[0]
            self.tests.setdefault(self.current_test, {})
        if TEST_REGEX["triggered"].match(line):
            self.tests[self.current_test]["triggered"] = self._triggered(
                self.current_test
            )
        if TEST_REGEX["started"].match(line):
            self.tests[self.current_test]["started"] = self._started(self.current_test)
        if TEST_REGEX["passed"].match(line):
            self.tests[self.current_test]["finished"] = self._finished(
                self.current_test, "PASSED"
            )
        if TEST_REGEX["failed"].match(line):
            self.tests[self.current_test]["finished"] = self._finished(
                self.current_test, "FAILED"
            )
        if TEST_REGEX["error"].match(line):
            self.tests[self.current_test]["finished"] = self._finished(
                self.current_test, "ERROR"
            )
        if TEST_REGEX["skipped"].match(line):
            self.tests[self.current_test]["finished"] = self._finished(
                self.current_test, "SKIPPED"
            )

    def execute(self):
        """Execute a test case."""
        self.logger.info("Figure out test directory.")
        test_directory = self.test_directory(self.checkout_command)
        line = False
        self.logger.info("Change directory to test directory '%s'.", test_directory)
        with self.etos.utils.chdir(test_directory):
            self.report_path = os.path.join(test_directory, self.report_path)
            self.logger.info("Report path: %s", self.report_path)
            self.logger.info("Executing pre-execution script.")
            self._pre_execution(self._build_environment_command())
            self.logger.info("Build test command")
            command = self._build_test_command()
            self.logger.info("Run test command: %s", command)
            iterator = self.etos.utils.iterable_call(
                [command], shell=True, executable="/bin/bash", output=self.report_path
            )
            self.logger.info("Start test.")
            for _, line in iterator:
                if TEST_REGEX:
                    self.parse(line)
            self.logger.info("Finished.")
            self.result = line



# Copyright 2020 Axis Communications AB.
#
# For a full list of individual contributors, please see the commit history.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""ETR test runner module."""
import time
import os
import logging
from pprint import pprint

from etos_test_runner.lib.graphql import request_test_suite_started
from etos_test_runner.lib.iut_monitoring import IutMonitoring
from etos_test_runner.lib.logs import LogHandler
from etos_test_runner.lib.executor import Executor

BASE = os.path.dirname(os.path.abspath(__file__))


class TestRunner:
    """Test runner for ETOS."""

    logger = logging.getLogger("ETR")

    def __init__(self, iut, etos):
        """Initialize.

        :param iut: IUT to execute tests on.
        :type iut: :obj:`etr.lib.iut.Iut`
        :param etos: ETOS library
        :type etos: :obj:`etos_lib.etos.ETOS`
        """
        self.etos = etos
        self.iut = iut
        self.config = self.etos.config.get("test_config")

        self.iut_monitoring = IutMonitoring(self.iut)
        self.issuer = {"name": "ETOS Test Runner"}

        self.env = os.environ
        if not os.path.isdir(os.getenv("TEST_ARTIFACT_PATH")):
            os.makedirs(os.getenv("TEST_ARTIFACT_PATH"))
        if not os.path.isdir(os.getenv("TEST_LOCAL_PATH")):
            os.makedirs(os.getenv("TEST_LOCAL_PATH"))
        if not os.path.isdir(os.getenv("GLOBAL_ARTIFACT_PATH")):
            os.makedirs(os.getenv("GLOBAL_ARTIFACT_PATH"))

    def test_suite_started(self):
        """Publish a test suite started event.

        :return: Reference to test suite started.
        :rtype: :obj:`eiffel.events.base_event.BaseEvent`
        """
        suite_name = self.config.get("name")
        categories = ["Regression test_suite", "Sub suite"]
        categories.append(self.iut.identity.name)
        livelogs = self.config.get("log_area", {}).get("livelogs")

        return self.etos.events.send_test_suite_started(
            suite_name,
            links={"CONTEXT": self.etos.config.get("context")},
            categories=categories,
            types=["FUNCTIONAL"],
            liveLogs=[{"name": "console", "uri": livelogs}],
        )

    def main_suite_id(self, activity_id):
        """Get the eiffel event id of the mainsuite linked form the activity triggered event.

        :param activity_id: Id of the activity linked to the main suite
        :type activity_id: str
        :return: Id of the main suite linked from the activity triggered event
        :rtype: str
        """
        for test_suite_started in request_test_suite_started(self.etos, activity_id):
            if not "Sub suite" in test_suite_started["data"]["testSuiteCategories"]:
                return test_suite_started["meta"]["id"]
        raise Exception("Missing main suite events, exiting!")

    def confidence_level(self, test_results, test_suite_started):
        """Publish confidence level modified event.

        :param test_results: Results of the test suite.
        :type test_results: bool
        :param test_suite_started: Test suite started event to use as Cause for this confidence.
        :type test_suite_started: :obj:`eiffellib.events.EiffelTestSuiteStartedEvent`
        """
        confidence = self.etos.events.send_confidence_level_modified(
            "{}_OK".format(self.config.get("name")),
            "SUCCESS" if test_results else "FAILURE",
            links={
                "CONTEXT": self.etos.config.get("context"),
                "CAUSE": test_suite_started,
                "SUBJECT": self.config.get("artifact"),
            },
            issuer=self.issuer,
        )
        print(confidence.pretty)

    def environment(self, context):
        """Send out which environment we're executing within.

        :param context: Context where this environment is used.
        :type context: str
        """
        # TODO: Get this from prepare
        if os.getenv("HOSTNAME") is not None:
            self.etos.events.send_environment_defined(
                "ETR Hostname",
                links={"CONTEXT": context},
                host={"name": os.getenv("HOSTNAME"), "user": "etos"},
            )

    def run_tests(self, log_handler):
        """Execute test recipes within a test executor.

        :param log_handler: Log handler for gathering logs.
        :type log_handler: :obj:`etr.lib.logs.LogHandler`
        :return: Result of test execution.
        :rtype: bool
        """
        recipes = self.config.get("recipes")
        result = True
        for num, test in enumerate(recipes):
            self.logger.info("Executing test %s/%s", num + 1, len(recipes))
            with Executor(test, self.iut, self.etos) as executor:
                self.logger.info("Starting test '%s'", executor.test_name)
                executor.execute()
                self.logger.info("Gather logs.")
                log_handler.gather_logs_for_executor(executor)

                if not executor.result:
                    result = executor.result
                self.logger.info("Test finished. Result: %s.", executor.result)
        return result

    def outcome(self, result, executed, description):
        """Get outcome from test execution.

        :param result: Result of execution.
        :type result: bool
        :param executed: Whether or not tests have successfully executed.
        :type executed: bool
        :param description: Optional description.
        :type description: str
        :return: Outcome of test execution.
        :rtype: dict
        """
        if executed:
            conclusion = "SUCCESSFUL"
            verdict = "PASSED" if result else "FAILED"
            self.logger.info(
                "Tests executed successfully. "
                "Verdict set to '%s' due to result being '%s'",
                verdict,
                result,
            )
        else:
            conclusion = "FAILED"
            verdict = "INCONCLUSIVE"
            self.logger.info(
                "Tests did not execute successfully. " "Setting verdict to '%s'",
                verdict,
            )

        suite_name = self.config.get("name")
        if not description and not result:
            self.logger.info(
                "No description but result is a failure. " "At least some tests failed."
            )
            description = "At least some {} tests failed.".format(suite_name)
        elif not description and result:
            self.logger.info(
                "No description and result is a success. "
                "All tests executed successfully."
            )
            description = "All {} tests completed successfully.".format(suite_name)
        else:
            self.logger.info("Description was set. Probably due to an exception.")
        return {
            "verdict": verdict,
            "description": description,
            "conclusion": conclusion,
        }

    def execute(self):  # pylint:disable=too-many-branches
        """Execute all tests in test suite.

        :return: Result of execution. Linux exit code.
        :rtype: int
        """
        self.logger.info("Send test suite started event.")
        test_suite_started = self.test_suite_started()
        sub_suite_id = test_suite_started.meta.event_id
        main_suite_id = self.main_suite_id(self.etos.config.get("context"))
        self.logger.info("Send test environment events.")
        self.environment(sub_suite_id)

        self.logger.info("Initialize loghandler.")
        log_handler = LogHandler(
            main_suite_id,
            sub_suite_id,
            self.etos.config.get("context"),
            self.iut,
            self.etos,
        )
        self.logger.info("Start IUT monitoring.")
        self.iut_monitoring.start_monitoring()

        result = True
        description = None
        try:
            self.logger.info("Starting test executor.")
            result = self.run_tests(log_handler)
            executed = True
        except Exception as exception:  # pylint:disable=broad-except
            result = False
            executed = False
            description = str(exception)
            raise
        finally:
            self.logger.info("Stop IUT monitoring.")
            self.iut_monitoring.stop_monitoring()

            self.logger.info("Gather global test logs.")
            log_handler.gather_global_logs()
            self.logger.info("Figure out test outcome.")
            outcome = self.outcome(result, executed, description)
            pprint(outcome)

            self.logger.info("Send test suite finished and confidence events.")
            self.etos.events.send_test_suite_finished(
                test_suite_started,
                links={"CONTEXT": self.etos.config.get("context")},
                outcome=outcome,
                persistentLogs=log_handler.persistent_logs,
            )
            self.confidence_level(result, test_suite_started)
        timeout = time.time() + 30
        self.logger.info("Waiting for eiffel publisher to deliver events (30s).")
        # pylint:disable=len-as-condition, protected-access
        while len(self.etos.publisher._deliveries):
            if time.time() > timeout:
                raise Exception("Eiffel publisher did not deliver all eiffel events.")
            time.sleep(1)
        self.logger.info("Tests finished executing.")
        return 0 if result else outcome



# Copyright 2020 Axis Communications AB.
#
# For a full list of individual contributors, please see the commit history.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Log handler test module."""
import os
import sys
import logging
import errno
from unittest import TestCase
from pathlib import Path
from mock import MagicMock, patch, call

from etos_test_runner.lib.logs import LogHandler

BASE_DIR = os.path.dirname(os.path.abspath(__file__))
logging.basicConfig(level=logging.DEBUG, stream=sys.stdout)
LOGGER = logging.getLogger(__name__)

# pylint:disable=protected-access


def step(msg):
    """Test step printer."""
    LOGGER.info("STEP: %s", msg)


class TestLogHandler(TestCase):
    """Tests for the ETR log handler."""

    def setUp(self):
        """Initialize a log cleanup list."""
        self.logs = []

    def tearDown(self):
        """Cleanup all log files created."""
        for log in self.logs:
            try:
                os.remove(log)
            except OSError:
                print("Failed to remove logfile %r" % log)

    @patch("etos_test_runner.lib.logs.os.rename")
    def test_move_log_to_pre_upload_path(self, mock_os_rename):
        """Test that it is possible to move a file to pre-upload path.

        Approval criteria:
            - Log handler shall create a new log file in the pre-upload path.

        Test steps::
            1. Initialize log handler.
            2. Create a log file to move.
            3. Tell log handler to move log file to pre-upload path.
            4. Verify that a new log file was created.
        """
        step("Initialize log handler.")
        log_handler = LogHandler("1", "2", "3", MagicMock(), MagicMock())

        step("Create a log file to move.")
        input_filename = Path(os.path.join(BASE_DIR, "a_pretty_log_name.log"))
        input_filename.touch()
        self.logs.append(input_filename)

        expected_filename = Path(
            os.path.join(log_handler.pre_upload_path, "a_new_log_name.log")
        )

        log = {"name": expected_filename.name, "file": str(input_filename)}
        step("Tell log handler to move log file to pre-upload path.")
        log_handler._move_log_to_pre_upload_path(log)

        step("Verify that a new log file was created.")
        mock_os_rename.assert_called_once_with(
            str(input_filename), str(expected_filename.absolute())
        )

    @patch("etos_test_runner.lib.logs.os.rename")
    def test_move_log_to_pre_upload_path_already_exists(self, mock_os_rename):
        """Test that log handler will rename the log file when a log file with name already exists.

        Approval criteria:
            - Log handler shall create a new log file in the pre-upload path.

        Test steps::
            1. Initialize log handler.
            2. Create a log file to move.
            3. Create a log file in the pre-upload path with the same name as log file.
            4. Tell log handler to move log file to pre-upload path.
            5. Verify that a new log file was created.
        """
        step("Initialize log handler.")
        log_handler = LogHandler("1", "2", "3", MagicMock(), MagicMock())

        step("Create a log file to move.")
        input_filename = Path(os.path.join(BASE_DIR, "a_pretty_log_name.log"))
        input_filename.touch()
        self.logs.append(input_filename)

        step("Create a log file in the pre-upload path with the same name as log file.")
        already_exists = Path(
            os.path.join(log_handler.pre_upload_path, "a_new_log_name.log")
        )
        already_exists.parent.mkdir(parents=True, exist_ok=True)
        already_exists.touch()
        self.logs.append(already_exists)

        expected_filename = Path(
            os.path.join(log_handler.pre_upload_path, "1_a_new_log_name.log")
        )

        log = {"name": already_exists.name, "file": str(input_filename)}
        step("Tell log handler to move log file to pre-upload path.")
        log_handler._move_log_to_pre_upload_path(log)

        step("Verify that a new log file was created.")
        mock_os_rename.assert_called_once_with(
            str(input_filename), str(expected_filename.absolute())
        )

    @patch("etos_test_runner.lib.logs.os.rename")
    def test_move_log_to_pre_upload_path_too_long_filename(self, mock_os_rename):
        """Test that log handler will reduce the length of logfile filename if it's too long.

        Approval criteria:
            - Log handler shall create a new log file in the pre-upload path.

        Test steps::
            1. Initialize log handler.
            2. Create a log file to move.
            3. Tell log handler to move log file to pre-upload path with a too long name.
            4. Verify that a new log file was created.
        """
        step("Initialize log handler.")
        log_handler = LogHandler("1", "2", "3", MagicMock(), MagicMock())
        exception = OSError()
        exception.errno = errno.ENAMETOOLONG
        mock_os_rename.side_effect = exception, True

        step("Create a log file to move.")
        input_filename = Path(os.path.join(BASE_DIR, "a_pretty_log_name.log"))
        input_filename.touch()
        self.logs.append(input_filename)

        expected_filename = Path(
            os.path.join(log_handler.pre_upload_path, "{}.log".format("a" * 249))
        )

        log = {"name": "{}.log".format("a" * 290), "file": str(input_filename)}
        step(
            "Tell log handler to move log file to pre-upload path with a too long name."
        )
        log_handler._move_log_to_pre_upload_path(log)
        calls = [
            call(
                str(input_filename),
                os.path.join(log_handler.pre_upload_path, "{}.log".format("a" * 290)),
            ),
            call(str(input_filename), str(expected_filename.absolute())),
        ]

        step("Verify that a new log file was created.")
        mock_os_rename.assert_has_calls(calls)



# Copyright 2020 Axis Communications AB.
#
# For a full list of individual contributors, please see the commit history.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""ETR log hander module."""
import os
import shutil
import logging
import traceback
import time
import errno
from json.decoder import JSONDecodeError
from copy import deepcopy
from pathlib import Path
from zipfile import ZipFile
from requests.auth import HTTPBasicAuth, HTTPDigestAuth
from requests.exceptions import HTTPError
from urllib3.exceptions import MaxRetryError, NewConnectionError


class LogHandler:  # pylint:disable=too-many-arguments,too-many-instance-attributes
    """Handle logfiles for ETOSTestRunner."""

    logger = logging.getLogger("LogHandler")

    def __init__(self, main_suite_id, sub_suite_id, context, iut, etos):
        """Initialize loghandler.

        :param main_suite_id: The id of the main suite.
        :type main_suite_id: str
        :param sub_suite_id:  The id of the sub suite.
        :type sub_suite_id: str
        :param context: Context in which this sub-suite is executing. Base path of logs.
        :type context: str
        :param iut: IUT this sub-suite is executing on.
        :type iut: :obj:`etr.lib.iut.Iut`
        :param etos: ETOS library instance.
        :type etos: :obj:`etos_lib.Etos`
        """
        self.logs = []
        self.iut = iut

        self.etos = etos
        self.context = context
        self.log_area = self.etos.config.get("test_config").get("log_area")

        # Path in log storage for the logs in this handler.
        self.log_storage_path = os.path.join(main_suite_id, sub_suite_id)

        # Storage location of logs and artifacts from test executions.
        self.executor_log_path = os.getenv("TEST_ARTIFACT_PATH")
        # Storage location of logs globally collected for ETR execution.
        self.global_log_path = os.getenv("GLOBAL_ARTIFACT_PATH")
        # Path for logs before uploading them to log storage area.
        self.pre_upload_path = os.getenv("TEST_LOCAL_PATH")

        self.full_execution_log_path = os.path.join(
            self.global_log_path, "full_execution.log"
        )

    @property
    def persistent_logs(self):
        """All persistent log formatted for EiffelTestSuiteFinishedEvent.

        :return: All persistent logs.
        :rtype: list
        """
        return [{"name": log["name"], "uri": log["uri"]} for log in self.logs]

    @property
    def executor_logs(self):
        """All logs found in executor log path."""
        for log in os.listdir(self.executor_log_path):
            yield log, os.path.join(self.executor_log_path, log)

    @property
    def global_logs(self):
        """All logs found in global log path."""
        for log in os.listdir(self.global_log_path):
            yield log, os.path.join(self.global_log_path, log)

    def _log_name_and_path(self, log, path, test_name=None):
        """Fix log names by attaching IUT identity and compress if log path is a folder.

        :param log: Name of log file to fix name on.
        :type log: str
        :param path: Path to logfile to fix name on.
        :type path: PathLike
        :param test_name: Name of test to which these logs belong.
        :type test_name: str
        :return: Name of log and path to log.
        :rtype: tuple
        """
        filename = log
        filepath = path

        if os.path.isdir(path):
            filepath, filename = self._compress_folder(path)

        # Attach test name to log.
        if test_name:
            filename = "{}_{}".format(test_name, filename)
        log_info = self.log_area.get("logs")
        if log_info:
            filename = "{}{}{}".format(
                log_info.get("prepend", ""),
                log_info.get("join_character", "_"),
                filename,
            )

        return filename, filepath

    @staticmethod
    def __auth(username, password, type="basic"):  # pylint:disable=redefined-builtin
        """Create an authentication for HTTP request.

        :param username: Username to authenticate.
        :type username: str
        :param password: Password to authenticate with.
        :type password: str
        :param type: Type of authentication. 'basic' or 'digest'.
        :type type: str
        :return: Authentication method.
        :rtype: :obj:`requests.auth`
        """
        if type.lower() == "basic":
            return HTTPBasicAuth(username, password)
        return HTTPDigestAuth(username, password)

    def retry_upload(
        self, verb, url, log_file, timeout=None, as_json=True, **requests_kwargs
    ):
        """Attempt to connect to url for x time.

        :param verb: Which HTTP verb to use. GET, PUT, POST
                     (DELETE omitted)
        :type verb: str
        :param url: URL to retry upload request
        :type url: str
        :param log_file: Opened log file to upload.
        :type log_file: file
        :param timeout: How long, in seconds, to retry request.
        :type timeout: int or None
        :param as_json: Whether or not to return json instead of response.
        :type as_json: bool
        :param request_kwargs: Keyword arguments for the requests command.
        :type request_kwargs: dict
        :return: HTTP response or json.
        :rtype: Response or dict
        """
        if timeout is None:
            timeout = self.etos.debug.default_http_timeout
        end_time = time.time() + timeout
        self.logger.debug(
            "Retrying URL %s for %d seconds with a %s request.", url, timeout, verb
        )
        iteration = 0
        while time.time() < end_time:
            iteration += 1
            self.logger.debug("Iteration: %d", iteration)
            try:
                # Seek back to the start of the file so that the uploaded file
                # is not 0 bytes in size.
                log_file.seek(0)
                yield self.etos.http.request(
                    verb, url, as_json, data=log_file, **requests_kwargs
                )
                break
            except (
                ConnectionError,
                HTTPError,
                NewConnectionError,
                MaxRetryError,
                TimeoutError,
                JSONDecodeError,
            ):
                self.logger.warning("%r", traceback.format_exc())
                time.sleep(2)
        else:
            raise ConnectionError(
                "Unable to {} {} with params {}".format(verb, url, requests_kwargs)
            )

    def _upload_log(self, context, log, name, folder):
        """Upload log to a storage location.

        :param context: Context for the http request.
        :type context: str
        :param log: Path to the log to upload.
        :type log: str
        :param name: Name of file to upload.
        :type name: str
        :param folder: Folder to upload to.
        :type folder: str
        :return: URI where log was uploaded to.
        :rtype: str
        """
        upload = deepcopy(self.log_area.get("upload"))
        data = {"context": context, "name": name, "folder": folder}

        # ETOS Library, for some reason, uses the key 'verb' instead of 'method'
        # for HTTP method.
        upload["verb"] = upload.pop("method")
        upload["url"] = upload["url"].format(**data)
        upload["timeout"] = upload.get("timeout", 30)
        if upload.get("auth"):
            upload["auth"] = self.__auth(**upload["auth"])

        with open(log, "rb") as log_file:
            for _ in range(3):
                request_generator = self.retry_upload(log_file=log_file, **upload)
                try:
                    for response in request_generator:
                        self.logger.debug("%r", response)
                        if not upload.get("as_json", True):
                            self.logger.debug("%r", response.text)
                        self.logger.info("Uploaded log %r.", log)
                        self.logger.info("Upload URI          %r", upload["url"])
                        self.logger.info("Data:               %r", data)
                        break
                    break
                except:  # noqa pylint:disable=bare-except
                    self.logger.error("%r", traceback.format_exc())
                    self.logger.error("Failed to upload log!")
                    self.logger.error("Attempted upload of %r", log)
        return upload["url"]

    def _rename_log_file_if_already_exists(self, target_path, log):
        """Rename a log file if it already exists.

        :param target_path: Which path to check if it exists.
        :type target_path: str
        :param log: Log to attempt to create.
        :type log: dict
        :return: New target path
        :rtype: str
        """
        index = 0
        while os.path.isfile(target_path):
            index += 1
            log["name"] = "{}_{}".format(index, log["name"])
            target_path = os.path.join(self.pre_upload_path, log["name"])
            self.logger.info(
                "Log with that name already exists. Rename to %r", target_path
            )
        return target_path

    def _move_log_to_pre_upload_path(self, log):
        """Move log file to pre upload path and rename if duplicates are found.

        :param log: Log to move.
        :type log: dict
        :return: Log with a new path and name.
        :rtype: dict
        """
        target_path = os.path.join(self.pre_upload_path, log["name"])
        self.logger.info("Moving log %r to %r", log, target_path)

        target_path = self._rename_log_file_if_already_exists(target_path, log)
        try:
            os.rename(log["file"], target_path)
        except OSError as error:
            if error.errno == errno.ENAMETOOLONG:
                target = Path(target_path)
                # Force the size of log name to 253 characters as to not crash due to filename
                # too long. 255 is, usually, the max size on Linux.
                target_path = os.path.join(
                    self.pre_upload_path,
                    target.name[: 253 - len(target.suffix)] + target.suffix,
                )
                self.logger.warning("Log name too long. Shortening to %r", target_path)
                target_path = self._rename_log_file_if_already_exists(target_path, log)
                os.rename(log["file"], target_path)
            else:
                raise
        log["file"] = target_path
        return log

    def gather_logs_for_executor(self, executor):
        """Gather and upload all logs for a single executor, connecting the logs to that execution.

        :param executor: Executor to gather logs from.
        :type executor: :obj:`etr.lib.executor.Executor`
        """
        with open(self.full_execution_log_path, "a+") as full_log:
            with open(executor.report_path) as report_log:
                full_log.write(report_log.read())

        for log, path in [*self.executor_logs, ("output.log", executor.report_path)]:
            filename, filepath = self._log_name_and_path(log, path, executor.test_name)
            if filename and filepath:
                log = self._move_log_to_pre_upload_path(
                    {
                        "file": filepath,
                        "name": filename,
                        "folder": self.log_storage_path,
                    }
                )
                log["uri"] = self._upload_log(
                    self.context, log["file"], log["name"], log["folder"]
                )
                self.logs.append(log)

    def gather_global_logs(self):
        """Gather and upload global ETR logs."""
        # Executor logs are fetched and stored into the global logs.
        # This is in order to try and fetch any stray logs after execution.
        for log, path in [*self.global_logs, *self.executor_logs]:
            filename, filepath = self._log_name_and_path(log, path)
            if filename and filepath:
                log = self._move_log_to_pre_upload_path(
                    {
                        "file": filepath,
                        "name": filename,
                        "folder": self.log_storage_path,
                    }
                )
                log["uri"] = self._upload_log(
                    self.context, log["file"], log["name"], log["folder"]
                )
                self.logs.append(log)

    @staticmethod
    def _compress_folder(absolute_folder_path):
        """Compress and remove folder from artifact path.

        Removal is done to avoid compressing the folder again
        and incrementally increasing the number of logs sent to
        log area or eiffel.

        :param absolute_folder_path: Path of folder to compress.
        :type absolute_folder_path: str
        :return: Tuple with zipfile path and filename.
        :rtype: tuple
        """
        zip_filename = "{}.zip".format(absolute_folder_path)
        with ZipFile(zip_filename, "w") as zip_file:
            for root, _, files in os.walk(absolute_folder_path):
                for _file in files:
                    filepath = os.path.relpath(os.path.join(root, _file))
                    zip_file.write(filepath)
        shutil.rmtree(absolute_folder_path, ignore_errors=True)
        return zip_filename, os.path.split(zip_filename)[1]



# Copyright 2020 Axis Communications AB.
#
# For a full list of individual contributors, please see the commit history.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""GraphAQL request handler module."""
from .graphql_queries import TEST_SUITE_STARTED


def request(etos, query):
    """Request graphql in a generator.

    :param etos: ETOS client instance.
    :type etos: :obj:`etos_lib.etos.Etos`
    :param query: Query to send to graphql.
    :type query: str
    :return: Generator
    :rtype: generator
    """
    wait_generator = etos.utils.wait(etos.graphql.execute, query=query)
    yield from wait_generator


def request_test_suite_started(etos, activity_id):
    """Request test suite started from graphql.

    :param etos: ETOS client instance.
    :type etos: :obj:`etos_lib.etos.Etos`
    :param activity_id: ID of activity in which the test suites started
    :type activity_id: str
    :return: Iterator of test suite started graphql responses.
    :rtype: iterator
    """
    for response in request(etos, TEST_SUITE_STARTED % activity_id):
        if response:
            for _, test_suite_started in etos.graphql.search_for_nodes(
                response, "testSuiteStarted"
            ):
                yield test_suite_started
            return None  # StopIteration
    return None  # StopIteration



#!/usr/bin/env python
# Copyright 2020 Axis Communications AB.
#
# For a full list of individual contributors, please see the commit history.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -*- coding: utf-8 -*-
"""ETOS test runner module."""
import argparse
import sys
import logging
import os
import shutil
import signal
from pprint import pprint

from etos_lib import ETOS

from etos_test_runner import __version__
from etos_test_runner.lib.testrunner import TestRunner
from etos_test_runner.lib.iut import Iut


# Remove spam from pika.
logging.getLogger("pika").setLevel(logging.WARNING)

_LOGGER = logging.getLogger(__name__)
BASE_DIR = os.path.dirname(os.path.abspath(__file__))
LOGFORMAT = "[%(asctime)s] %(levelname)s:%(message)s"
logging.basicConfig(
    level=logging.DEBUG,
    stream=sys.stdout,
    format=LOGFORMAT,
    datefmt="%Y-%m-%d %H:%M:%S",
)


def parse_args(args):
    """Parse command line parameters.

    :param args: command line parameters as list of strings
    :return: command line parameters as :obj:`airgparse.Namespace`
    """
    parser = argparse.ArgumentParser(description="ETOS test runner")
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version="etos_test_runner {ver}".format(ver=__version__),
    )
    return parser.parse_args(args)


class ETR:
    """ETOS Test Runner."""

    context = None

    def __init__(self):
        """Initialize ETOS library and start eiffel publisher."""
        self.etos = ETOS("ETOS Test Runner", os.getenv("HOSTNAME"), "ETOS Test Runner")
        self.etos.config.rabbitmq_publisher_from_environment()
        # ETR will print the entire environment just before executing.
        # Hide the password.
        os.environ["RABBITMQ_PASSWORD"] = "*********"

        self.etos.start_publisher()
        self.tests_url = os.getenv("SUB_SUITE_URL")
        os.environ["GLOBAL_ARTIFACT_PATH"] = os.getenv(
            "GLOBAL_ARTIFACT_PATH", "/home/etos/global"
        )
        os.environ["TEST_ARTIFACT_PATH"] = os.getenv(
            "TEST_ARTIFACT_PATH", "/home/etos/artifacts"
        )
        os.environ["TEST_LOCAL_PATH"] = os.getenv("TEST_LOCAL_PATH", "/home/etos/local")
        signal.signal(signal.SIGTERM, self.graceful_shutdown)

    @staticmethod
    def graceful_shutdown(*args):
        """Catch sigterm."""
        raise Exception("ETR has been terminated.")

    def download_and_load(self):
        """Download and load test json."""
        generator = self.etos.http.wait_for_request(self.tests_url)
        for response in generator:
            json_config = response
            break
        self.etos.config.set("test_config", json_config)
        self.etos.config.set("context", json_config.get("context"))

    def _run_tests(self):
        """Run tests in ETOS test runner.

        :return: Results of test runner execution.
        :rtype: bool
        """
        iut = Iut(self.etos.config.get("test_config").get("iut"))
        test_runner = TestRunner(iut, self.etos)
        return test_runner.execute()

    @staticmethod
    def _clear_artifacts():
        """Clear old artifacts from file system."""
        artifact_path = os.getenv("TEST_ARTIFACT_PATH")
        local_path = os.getenv("TEST_LOCAL_PATH")
        if os.path.exists(artifact_path):
            shutil.rmtree(artifact_path, ignore_errors=True)
        if os.path.exists(local_path):
            shutil.rmtree(local_path, ignore_errors=True)

    def run_etr(self):
        """Send activity events and run ETR.

        :return: Result of testrunner execution.
        :rtype: bool
        """
        _LOGGER.info("Starting ETR.")
        self.download_and_load()
        try:
            activity_name = self.etos.config.get("test_config").get("name")
            triggered = self.etos.events.send_activity_triggered(activity_name)
            self._clear_artifacts()
            top_dir = os.getcwd()

            with self.etos.utils.chdir(top_dir):
                self.etos.events.send_activity_started(triggered)
                result = self._run_tests()
        except Exception as exc:  # pylint:disable=broad-except
            self.etos.events.send_activity_finished(
                triggered, {"conclusion": "FAILED", "description": str(exc)}
            )
            raise
        self.etos.events.send_activity_finished(triggered, {"conclusion": "SUCCESSFUL"})
        _LOGGER.info("ETR finished.")
        return result


def main(args):
    """Start ETR."""
    args = parse_args(args)

    etr = ETR()
    result = etr.run_etr()
    if isinstance(result, dict):
        pprint(result)
    _LOGGER.info("Done. Exiting")
    _LOGGER.info(result)
    sys.exit(result)


def run():
    """Entry point to ETR."""
    logging.basicConfig(level=logging.INFO, stream=sys.stdout)
    main(sys.argv[1:])


if __name__ == "__main__":
    run()
